#include "../config.h"
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <string.h>
#include "libnjb.h"
#include "njbusb.h"
#include "protocol.h"
#include "protocol3.h"
#include "unicode.h"
#include "byteorder.h"
#include "njb_error.h"
#include "usb_io.h"
#include "ioutil.h"
#include "defs.h"
#include "base.h"

extern int __sub_depth;

/*
 * NJB2,3,Zen,Zen USB 2.0, Zen NX and Zen Xtra 
 * specific functions in the protocol goes 
 * in here. This is a work in progress, not all of the NJB3-style 
 * protocol instructions are yet implemented.
 */

static char *int32tostring(u_int32_t val)
{
  char tmp[40];
  sprintf(tmp,"%u", val);
  return strdup(tmp);
}



static int send_njb3_command(njb_t *njb, unsigned char *command, int clength)
{
  __dsub= "send_njb3_command";
  
  unsigned char *data;

  /* The header differs for NJB2 and NJB Zen USB 2.0 */
  unsigned char usb11_cmd_header[]={0x43,0x42,0x53,0x55,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,
				   0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0};
  unsigned char usb20_cmd_header[]={0x55,0x53,0x42,0x43,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0, 0x0,0x0,0x0,0x0,
				   0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0,0x0};
  ssize_t cmd_len;

  /* Structure of "command follows" command: 0x20 bytes
   * 4 bytes unknown 0x43 0x42 0x53 0x55 comprising the string "CBSU", for NJB2 "USBC" (byte order??)
   * 6 bytes unknown 0x00 * 6
   * 2 bytes total command length (or is it 4 bytes? or 8?)
   * 20 bytes unknown 0x00 * 20
   */
  
  __enter;
  
  data= (unsigned char *) malloc (0x20);
  if ( data == NULL ) {
    NJB_ERROR(EO_NOMEM);
    __leave;
    return -1;
  }

  /* Use the apropriate header on USB 2.0 devices as
   * opposed to USB 1.1 devices. */
  if (USB20_DEVICE(njb->device_type)) {
    memcpy(data, usb20_cmd_header, 0x20);
    cmd_len = 0x1F;
  } else {
    memcpy(data, usb11_cmd_header, 0x20);
    cmd_len = 0x20;
  }

  /* Set command length parameter */
  from_16bit_to_njb3_bytes(clength, &data[10]);
  
  if (usb_pipe_write(njb, data, cmd_len) == -1) {
    free(data);
    NJB_ERROR(EO_USBBLK);
    __leave;
    return -1;
  }
  
  free(data);

  if (usb_pipe_write(njb, command, clength) == -1) {
    NJB_ERROR(EO_USBBLK);
    __leave;
    return -1;
  }
  
  __leave;
  return 0;
}

int njb3_capture (njb_t *njb)
{
	__dsub= "njb3_capture";

	__enter;

       	if ( usb_setup(&njb->bin, UT_CLASS,
       		NJB_CMD_CAPTURE_NJB3, 0, 0, 0, NULL) == -1 ) {
       		NJB_ERROR(EO_USBBLK);
		goto errcap;
       	}

	__leave;
	return 0;

 errcap:
	usb_setup(&njb->ctl, UT_CLASS, 
		  NJB_CMD_RELEASE_NJB3, 0, 0, 0, NULL);
	__leave;
	return -1;
}


int njb3_release(njb_t *njb)
{
	__dsub= "njb3_capture";

	__enter;

	if (usb_setup(&njb->ctl, UT_CLASS, 
		      NJB_CMD_RELEASE_NJB3, 0, 0, 0, NULL) == -1) {
	  goto errls;
	}
		
	__leave;
	return 0;

 errls:
	__leave;
	return -1;
}

/* Read a sequence of unknown use */
int njb3_read_unknown_1(njb_t *njb)
{
	__dsub= "njb3_read_unknown_1";
	ssize_t bread;
	unsigned char data[14];
       	unsigned char njb3_read_unknown_1[]={0x00,0x08,0x00,0x01,0xff,0xfe,0x00,0x02,0x00,0x01,0x00,0x00};
	__enter;

       	/* Retrieves 8 bytes of unknown use */
       	if (send_njb3_command(njb, njb3_read_unknown_1, 0x0c) == -1){
		goto err1;
       	}
       	if ( (bread= usb_pipe_read(njb, data, 14)) == -1 ) {
       		NJB_ERROR(EO_USBBLK);
		printf("bread1:%d\n", bread);
		goto err1;
       	} else if ( bread < 14 ) {
       		NJB_ERROR(EO_RDSHORT);
		printf("bread2:%d\n", bread);
		goto err1;
       	}
	return 0;

err1:
	__leave;
	return -1;
}

/* We don't know the use for this one either */
int njb3_read_ar00pl00sg00(njb_t *njb)
{
	__dsub= "njb3_read_ar00pl00sg00";
	ssize_t bread;
	unsigned char data[0x9a];
       	unsigned char njb3_read_ar00pl00sg00[]={0x00,0x0c,0x00,0x01,0x00,0x00,0x00,0x0a,
						0x14,0x00,0x00,0x06,0x00,0x0a,0x00,0x14,
						0x00,0x15,0x00,0x00};
	__enter;

       	if (send_njb3_command(njb, njb3_read_ar00pl00sg00, 20) == -1){
		goto err1;
       	}
       	if ( (bread= usb_pipe_read(njb, data, 0x9a)) == -1 ) {
       		NJB_ERROR(EO_USBBLK);
		printf("bread1:%d\n", bread);
		goto err1;
       	} else if ( bread < 0x9a ) {
       		NJB_ERROR(EO_RDSHORT);
		printf("bread2:%d\n", bread);
		goto err1;
       	}

	/* Create a standalone parse file for this, like for
	 * songid etc. */

	return 0;

err1:
	__leave;
	return -1;
}

int njb3_ping (njb_t *njb, njbid_t *njbid, int type)
{
	__dsub= "njb3_ping";
	int rc = -1;
	ssize_t bread;
       	unsigned char data[256];
	/*                                            -0x00 0x80- on first call           */
	/*                                            -0xff 0xfe- on subsequent calls...? */
       	unsigned char njb3_ping0[]={0x00,0x08,0x00,0x01,0xff,0xfe,0x00,0x02,0x00,0x03,0x00,0x00,0x00,0x00};
       	unsigned char njb3_ping1[]={0x00,0x08,0x00,0x01,0xff,0xfe,0x00,0x02,0x00,0x03,0x00,0x00};

       	/* Structure
       	 * 2 bytes unknown 0x0008 (could mean "read operation")
       	 * 2 bytes unknown 0x0001
       	 * 2 bytes unknown 0x0080
       	 * 2 bytes request code length 0x0002
       	 * 2 bytes metadata frame name 0x0003 = Product ID
       	 * 2 bytes termination 0x0000
       	 */
       	char *tmp;

	__enter;


	/* Some special handling for USB 2.0 devices, perhaps this should be 
	 * used for NJB3/ZEN too?
	 */
	if (USB20_DEVICE(njb->device_type)) {
	  if (njb3_capture(njb) == -1)
	    goto err1;
	}
	/* Here we used to retrieve the play status - useless. */

       	/* Retrieve the device identification, 52 bytes for NJB2/3
	 * 56 bytes for NJB Zen / Zen 2.0 */
	if (type == 0) {
	    if (send_njb3_command(njb, njb3_ping0, 0x0c) == -1)
		    goto err1;
	} else {
	    if (send_njb3_command(njb, njb3_ping1, 0x0c) == -1)
		    goto err1;
       	}

	/* We don't check for the length of this */
	if ( (bread= usb_pipe_read(njb, data, 256)) == -1 ) {
	  NJB_ERROR(EO_USBBLK);
	  goto err1;
	}

	/* Special NJB2 fix from Berni to release/capture NJB2. Is this applicable
	 * for NJB3/ZEN too? We applied it to all USB 2.0 devices. */
	if (USB20_DEVICE(njb->device_type)) {
	  if (njb3_release(njb) == -1)
	    goto err1;
	  if (njb3_capture(njb) == -1)
	    goto err1;
	}
	
       	if ( data[0] ) {
       		NJB_STATUS(data[0]);
		goto err1;
       	}

	if (njbid) {
	  njbid->fwRel= (u_int8_t) data[11];
	  njbid->fwMinor= (u_int8_t) data[9];
	  njbid->fwMajor= (u_int8_t) data[7];
	  /* Copy this -- eventually move it into njb_t */
	  njb->fwRel = njbid->fwRel;
	  njb->fwMinor = njbid->fwMinor;
	  njb->fwMajor = njbid->fwMajor;
	  if ((tmp = ucs2tostr(&data[18])) != NULL) {
	    memcpy(&njbid->productName, tmp, strlen(tmp)+1);
	    free(tmp);
	  }
	  /* The "power connected" information seems not to exist in the 
	   * NJB3 ping data, perhaps this must be retrieved by a separate
	   * command... */
	  /* njbid->power= (u_int8_t) data[13]; */
	}

	rc = 0;
err1:
	__leave;
	return rc;
}

int njb3_readid (njb_t *njb, njbid_t *njbid)
{
	__dsub= "njb3_readid";
	int rc = -1;
	ssize_t bread;
	unsigned char data[24];
	/*                                              -0x00 0x80- on first call            */
	/*                                              -0xff 0xfe- on subsequent calls...?  */
       	unsigned char njb3_readid[]={0x00,0x08,0x00,0x01,0xff,0xfe,0x00,0x02,0x00,0x15,0x00,0x00};
       	/* Structure
       	 * 2 bytes unknown 0x0008 (could mean "read operation")
       	 * 2 bytes unknown 0x0001
       	 * 2 bytes unknown 0x0080
       	 * 2 bytes request code length 0x0002
       	 * 2 bytes metadata frame name 0x0015 = NJB device ID, 16 bytes
       	 * 2 bytes termination 0x0000
       	 */

	__enter;

       	/* This retrieves the jukebox ID */
       	if (send_njb3_command(njb, njb3_readid, 0x0c) == -1){
		goto err1;
       	}
       	if ( (bread= usb_pipe_read(njb, data, 24)) == -1 ) {
       		NJB_ERROR(EO_USBBLK);
		printf("bread1:%d\n", bread);
		goto err1;
       	} else if ( bread < 24 ) {
       		NJB_ERROR(EO_RDSHORT);
		printf("bread2:%d\n", bread);
		goto err1;
       	}

       	memcpy(&njbid->id, &data[6], 16);

	rc = 0;
err1:
	__leave;
	return rc;
}


int njb3_get_disk_usage (njb_t *njb, u_int64_t *totalbytes, u_int64_t *freebytes)
{
	__dsub= "njb3_get_disk_usage";
	u_int32_t lsdw, msdw;
	unsigned char data[0x14];

	unsigned char get_files_and_dirs[]={0x00,0x08,0x00,0x01,0xff,0xfe,0x00,0x02,0x00,0x13,0x00,0x00};
       	/* Structure
       	 * 2 bytes unknown 0x0008 (could mean "read operation")
       	 * 2 bytes unknown 0x0001
       	 * 2 bytes unknown 0xfffe
       	 * 2 bytes request code length 0x0002
       	 * 2 bytes metadata frame name 0x0013
       	 * 2 bytes termination 0x0000
       	 */
	unsigned char get_disk_usage[]={0x00,0x08,0x00,0x01,0xff,0xfe,0x00,0x02,0x00,0x02,0x00,0x00};
       	/* Structure
       	 * 2 bytes unknown 0x0008 (could mean "read operation")
       	 * 2 bytes unknown 0x0001
       	 * 2 bytes unknown 0xfffe
       	 * 2 bytes request code length 0x0002
       	 * 2 bytes metadata frame name 0x0002
       	 * 2 bytes termination 0x0000
       	 */

	ssize_t bread;

	__enter;

	if (send_njb3_command(njb, get_disk_usage, 0x0c) == -1) {
		__leave;
		return -1;
	}
	if ( (bread= usb_pipe_read(njb, data, 0x14)) == -1 ) {
		NJB_ERROR(EO_USBBLK);
		__leave;
		return -1;
	} else if ( bread < 0x14 ) {
		NJB_ERROR(EO_RDSHORT);
		__leave;
		return -1;
	}

	msdw = 0;

	lsdw = njb3_bytes_to_32bit(&data[0x0a]);
	*totalbytes=  make64(msdw, lsdw) * 1024;
	lsdw = njb3_bytes_to_32bit(&data[0x0e]);
	*freebytes=  make64(msdw, lsdw) * 1024;

	/* FIXME: This command returns the files and playlist counts etc, it is
	 *        currently not used for anything useful.
	 */
	if (send_njb3_command(njb, get_files_and_dirs, 0x0c) == -1) {
		__leave;
		return -1;
	}
	if ( (bread= usb_pipe_read(njb, data, 0x0c)) == -1 ) {
		NJB_ERROR(EO_USBBLK);
		__leave;
		return -1;
	} else if ( bread < 0x0c ) {
		NJB_ERROR(EO_RDSHORT);
		__leave;
		return -1;
	}
	
        __leave;
        return 0;
}

// This is probably the way to retrieve singe string frames. 
// Short reads on a large buffer. Currently only used for
// getting the owner string.
static char* njb3_read_string_frame(njb_t *njb, u_int16_t frameid)
{
  __dsub= "njb3_read_string_frame";
  unsigned char get_frame[]={0x00,0x08,0x00,0x01,0xff,0xfe,0x00,0x02,0x00,0x00,0x00,0x00};
  /* Structure
   * 2 bytes unknown 0x0008 (could mean "read string operation")
   * 2 bytes unknown 0x0001
   * 2 bytes unknown 0xfffe - magic numbers... first 0x0080 then 0xfffe?
   * 2 bytes request code length 0x0002
   * 2 bytes metadata frame name (supplied as parameter)
   * 2 bytes termination 0x0000
   *
   * returned structure is:
   * 2 bytes status 0x0000
   * 2 bytes length, in bytes of return
   * 2 bytes frame ID
   * n bytes of data, usually terminated by 0000
   */
  ssize_t bread;
  // Reading buffer
  unsigned char data[NJB3_SHORTREAD_BUFSIZE];
  u_int16_t status;
  u_int16_t strlen;
  char *tmpstr;

  __enter;

  from_16bit_to_njb3_bytes(frameid, &get_frame[8]);

  if (send_njb3_command(njb, get_frame, 0x0c) == -1){
    __leave;
    return NULL;
  }

  // fprintf(stdout, 
  // "Reading varlen string frame 0x%04x max 1024 bytes...\n", frameid);

  if ( (bread= usb_pipe_read(njb, data, NJB3_SHORTREAD_BUFSIZE)) == -1 ) {
    NJB_ERROR(EO_USBBLK);
    __leave;
    return NULL;
  } else if ( bread < 6 ) {
    // We need atleast the string length!
    NJB_ERROR(EO_RDSHORT);
    __leave;
    return NULL;
  }

  // I think this is the status value.
  status = njb3_bytes_to_16bit(&data[0]);
  strlen = njb3_bytes_to_16bit(&data[2]);
  if (status != 0) {
    // Not actually this error but what have I...
    NJB_ERROR(EO_RDSHORT);
    __leave;
    return NULL;
  }

  // Extract string value
  if (strlen == 0) {
    __leave;
    return NULL;
  }  
  tmpstr = ucs2tostr(&data[6]);
  
  __leave;
  return tmpstr;
}

int njb3_get_owner_string (njb_t *njb, char *name)
{
	__dsub= "njb3_get_owner_string";
       	char *tmp;

	__enter;

	tmp = njb3_read_string_frame(njb, NJB3_OWNER_FRAME_ID);
	if (tmp == NULL) {
	  __leave;
	  return -1;
	}

	/* The destination field is no bigger than OWNER_STRING_LENGTH
	 * so this is the practical limit. The result if ucs2tostr() may
	 * be > 100 because of Unicode characters as UTF-8 */
	strncpy(name, tmp, OWNER_STRING_LENGTH);
	free(tmp);
	name[OWNER_STRING_LENGTH]= '\0'; /* Just in case */
	__leave;
	return 0;
}

int njb3_set_owner_string (njb_t *njb, const char *name)
{
	__dsub= "njb3_set_owner_string";
       	unsigned char njb3_set_owner[]={0x00,0x07,0x00,0x01,0x00,0x00,0x01,0x13};
       	/* TODO: this seems to overwrite just as many characters there are in
	 * the supplied string argument, perhaps it has to be padded with trailing
	 * spaces up to 100 bytes (max string length)?
	 *
	 * Structure: 8 bytes
       	 * 2 bytes unknown 0x0007 (could mean "write operation")
       	 * 2 bytes unknown 0x0001
       	 * 2 bytes string length
       	 * 2 bytes owner string tag 0x0113
	 * then follows the n bytes of the owner string.
       	 */
       	unsigned char *unistr;
       	unsigned char *data;
	unsigned char status[2];
       	int unilen;
       	int cmdlen;
	int bread;

	__enter;

       	/* 8 bytes header
       	 * Unicode string length + extra 0x00 0x00
       	 */
       	unistr = strtoucs2(name);
       	unilen = ucs2strlen(unistr)*2;
       	cmdlen = 8 + unilen + 4;

       	data= (unsigned char *) malloc (cmdlen);
       	if ( data == NULL ) {
       		NJB_ERROR(EO_NOMEM);
       		__leave;
       		return -1;
       	}
       	memset(data, 0, cmdlen);
		
       	memcpy(&data[0], njb3_set_owner, 0x08);
       	memcpy(&data[8], unistr, unilen+2);
	from_16bit_to_njb3_bytes(unilen+4, &data[4]);

       	if (send_njb3_command(njb, data, cmdlen) == -1){
       		free(data);
       		__leave;
       		return -1;
       	}

       	free(data);

	/* Then read two confirmation bytes back as status, only
	 * verified for NJB Zen USB 2.0 */
	if ( (bread= usb_pipe_read (njb, &status, 2)) == -1 ) {
	  NJB_ERROR (EO_USBBLK);
	  __leave;
	  return -1;
	}
	else if (bread < 2) {
	  NJB_ERROR (EO_RDSHORT);
	  __leave;
	  return -1;
	}
	if (njb3_bytes_to_16bit(&status[0]) != 0) {
	  __leave;
	  return -1;
	}

	__leave;
	return 0;
}

njb_time_t *njb3_get_time(njb_t *njb)
{
	__dsub= "njb3_get_time";
       	unsigned char njb3_get_time[]={0x00,0x08,0x00,0x01,0xff,0xfe,0x00,0x02,0x01,0x10,0x00,0x00};
	/* Structure: 8 bytes
       	 * 2 bytes unknown 0x0008 (could mean "read operation")
       	 * 2 bytes unknown 0x0001
       	 * 2 bytes unknown 0xfffe
       	 * 2 bytes request code length 0x0002
       	 * 2 bytes metadata frame name 0x0110
       	 * 2 bytes termination 0x0000
       	 */
	njb_time_t *time = NULL;
	unsigned char *data;
	ssize_t bread;


	__enter;

	data= malloc(16);
	if (data == NULL) {
		NJB_ERROR(EO_NOMEM);
		__leave;
		return NULL;
	}

       	if (send_njb3_command(njb, njb3_get_time, 0x0c) == -1){
       		__leave;
       		return NULL;
       	}
       	if ( (bread= usb_pipe_read(njb, data, 16)) == -1 ) {
       		NJB_ERROR(EO_USBBLK);
       		__leave;
       		return NULL;
       	}

	/* NOT USED? SHORT READ? This need to be investigated.
	  else if ( bread < 16 ) {
	  NJB_ERROR(EO_RDSHORT);
	  __leave;
	  return NULL;
	  }
	*/

       	time= time_unpack3(data);
	
	free(data);

	__leave;
	return time;
}

/* Set up the black/white bitmap shown on shutdown.
 *
 * The bitmap must be exactly 1088 bytes large,
 * coded as a serial bitmap of 136x64 pixels. The
 * final 4 bits per line are not shown, because the
 * the display of the NJB2 has only 132x64 pixels.
 *
 * A set pixel (1) means dark, an unset means white.
 *
 * To create a compatible bitmap, take a 132x64 PBM file
 * (of course not a ASCII-coded, but a binary "P4") and
 * cut off the header.
 */
int njb2_set_bitmap(njb_t *njb, unsigned char * bitmap)
{
  __dsub= "njb2_set_bitmap";

  unsigned char announce_image[0x0A] = {
    0x00, 0x0b, 0x00, 0x01, 0x00, 0x02, 0x00, 0x00,
    0x04, 0x2c
  };
  /* Structure:
   * 2 bytes unknown 0x000b
   * 2 bytes unknown 0x0001
   * 2 bytes unknown 0x0002
   * 2 bytes unknown 0x0000
   * 2 bytes image data length 0x042c
   */
  unsigned char image_data[0x42C] = {
    0x4a, 0x42, 0x4d, 0x31, 0x00, 0x84, 0x00, 0x40,
    0x00, 0x00, 0x00, 0x01 // write the transcoded bitmap here
  };
  /* Image header:
   * 4 bytes signature "JBM1"
   * 2 bytes width  0x0084 = 132
   * 2 bytes height 0x0040 =  64
   * 2 bytes unknown 0x0000
   * 2 bytes unknown 0x0001
   *
   * Maybe the last two words of the image header are used as
   * palette index or something.
   */
  /* The real image data starts at ... */
  unsigned char * image = image_data + 12;
  ssize_t bread;
  unsigned char result[2];
  int x,y;

  __enter;

  /* Transcode the bitmap.
   *
   * The NJB2 uses a "packed" bitmap format, therefore the serial
   * bitmap has to be transcoded:
   * The NJB2-bitmap-format is 16bit-word based. Every word defines
   * the pixels of a 2x8 area, starting with the most-significant-bit
   * in the lower left corner and ending with the lsb in the upper
   * right corner.
   * The bitmap consists of 528 of these areas. The first word defines
   * the area in the upper left corner.
   */

  /* For better readable code ;) I broke up the transcoding into columns
   * and rows. Note that a I am talking of columns and rows of 2x8 subareas.
   *
   * Instead of 16bit-word bases transcoding 8bit samples are used (that
   * means the two columns of the 2x8 area are transcoded individually).
   */
  for (y = 0; y < 8; y++) {
    for (x = 0; x < 66; x++) {
      u_int8_t low_byte_mask  = 0x80 >> (2*(x%4));
      u_int8_t high_byte_mask = 0x40 >> (2*(x%4));

      image[y*132+x*2]
	= ((bitmap[y*136+0*17+x/4] & (high_byte_mask)) ? 0x01 : 0x00)
	| ((bitmap[y*136+1*17+x/4] & (high_byte_mask)) ? 0x02 : 0x00)
	| ((bitmap[y*136+2*17+x/4] & (high_byte_mask)) ? 0x04 : 0x00)
	| ((bitmap[y*136+3*17+x/4] & (high_byte_mask)) ? 0x08 : 0x00)
	| ((bitmap[y*136+4*17+x/4] & (high_byte_mask)) ? 0x10 : 0x00)
	| ((bitmap[y*136+5*17+x/4] & (high_byte_mask)) ? 0x20 : 0x00)
	| ((bitmap[y*136+6*17+x/4] & (high_byte_mask)) ? 0x40 : 0x00)
	| ((bitmap[y*136+7*17+x/4] & (high_byte_mask)) ? 0x80 : 0x00);
      image[y*132+x*2+1]
	= ((bitmap[y*136+0*17+x/4] & (low_byte_mask)) ? 0x01 : 0x00)
	| ((bitmap[y*136+1*17+x/4] & (low_byte_mask)) ? 0x02 : 0x00)
	| ((bitmap[y*136+2*17+x/4] & (low_byte_mask)) ? 0x04 : 0x00)
	| ((bitmap[y*136+3*17+x/4] & (low_byte_mask)) ? 0x08 : 0x00)
	| ((bitmap[y*136+4*17+x/4] & (low_byte_mask)) ? 0x10 : 0x00)
	| ((bitmap[y*136+5*17+x/4] & (low_byte_mask)) ? 0x20 : 0x00)
	| ((bitmap[y*136+6*17+x/4] & (low_byte_mask)) ? 0x40 : 0x00)
	| ((bitmap[y*136+7*17+x/4] & (low_byte_mask)) ? 0x80 : 0x00);
    }
  }

  if (send_njb3_command (njb, announce_image, 0x0a) == -1) {
    __leave;
    return -1;
  }

  if (send_njb3_command (njb, image_data, 0x042c) == -1) {
    __leave;
    return -1;
  }

  if ( (bread= usb_pipe_read (njb, &result, 2)) == -1 ) {
    NJB_ERROR (EO_USBBLK);
    __leave;
    return -1;
  }
  else if (bread < 2) {
    NJB_ERROR (EO_RDSHORT);
    __leave;
    return -1;
  }

  __leave;
  return 256*result[0] + result[1];
}

int njb3_current_track (njb_t *njb, u_int16_t *position)
{
  __dsub= "njb3_current_track";

  unsigned char njb3_query [0x0c] = {
    0x00, 0x08, 0x00, 0x01, 0xff, 0xfe, 0x00, 0x02,
    0x01, 0x19, 0x00, 0x00
  };
  unsigned char status[0x0a];
  ssize_t bread;

  __enter;

  if (send_njb3_command (njb, njb3_query, 0x0c) == -1) {
    __leave;
    return -1;
  }

  if ( (bread= usb_pipe_read (njb, status, 0x0a)) == -1 ) {
    NJB_ERROR (EO_USBBLK);
    __leave;
    return -1;
  }
  else if (bread < 0x0a) {
    NJB_ERROR (EO_RDSHORT);
    __leave;
    return -1;
  }

  *position = (u_int16_t) (status[8]) << 8 | status[7];

  __leave;
  return 0;
}

int njb3_elapsed_time(njb_t *njb, u_int16_t * elapsed, int * change)
{
  __dsub= "njb3_elapsed_time";

  unsigned char get_elapsed_time [0x04] = { 0x01, 0x01, 0x00, 0x01 };
  unsigned char data[4];
  ssize_t bread;

  __enter;

  if (send_njb3_command (njb, get_elapsed_time, 0x04) == -1) {
    __leave;
    return -1;
  }

  if ( (bread= usb_pipe_read (njb, data, 4)) == -1 ) {
    NJB_ERROR (EO_USBBLK);
    __leave;
    return -1;
  }
  else if (bread < 4) {
    NJB_ERROR (EO_RDSHORT);
    __leave;
    return -1;
  }

  *elapsed = 256*data[2] + data[3];

  /* If elapsed is relatively small, check if
   * a new track is played.
   *
   * If elapsed_time () is called less frequently (than 2 secs),
   * a change may not be recognized! */
  if (*elapsed < 2) {
    u_int16_t track;

    if (njb3_current_track (njb, &track) == -1) {
      __leave;
      return -1;
    }

    if (track != njb->current_playing_track) {
      njb->current_playing_track = track;
      *change = 1;
    }
  }

  __leave;

  return 0;
}


int njb3_queue_track(njb_t *njb, u_int32_t trackid)
{
  __dsub= "njb3_queue_track";

  ssize_t bread;
  unsigned char njb3_enqueue_track[0x10] = {
    0x01, 0x04, 0x00, 0x01, 0xFF, 0xFF, 0x00, 0x06,
    0x01, 0x1c, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
  };
  /* Structure njb3_enqueue_track: 16 bytes
   * 2 bytes unknown 0x0104
   * 2 bytes unknown 0x0001
   * 2 bytes unknown 0xFFFF
   * 2 bytes request code length 0x0006
   * 2 bytes metadata frame name 0x011c
   * 4 bytes track id
   * 2 bytes termination 0x0000
   */
  unsigned char data[2];

  __enter;

  njb3_enqueue_track[0xA] = (trackid >> 24) & 255;
  njb3_enqueue_track[0xB] = (trackid >> 16) & 255;
  njb3_enqueue_track[0xC] = (trackid >> 8) & 255;
  njb3_enqueue_track[0xD] = trackid & 255;

  if (send_njb3_command (njb, njb3_enqueue_track, 0x10) == -1) {
    __leave;
    return -1;
  }

  if ( (bread= usb_pipe_read(njb, data, 2)) == -1 ) {
    NJB_ERROR(EO_USBBLK);
    __leave;
    return -1;
  }  
  else if (bread < 2) {
    NJB_ERROR (EO_RDSHORT);
    __leave;
    return -1;
  }

  __leave;
  return 0;
}

int njb3_play_track(njb_t *njb, u_int32_t trackid)
{
  __dsub= "njb3_play_track";

  ssize_t bread;
  unsigned char njb3_start_play [0x10] = {
    0x01, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x04, 0x01, 0x19, 0x00, 0x00, 0x00, 0x00
  };
  /* Structure njb3_start_play: 16 bytes
   * 2 bytes unknown 0x0100
   * 2 bytes unknown 0x0001
   * 4 bytes track id
   * 2 bytes request code length 0x0004
   * 2 bytes metadata frame name 0x0119
   * 2 bytes unknown 0x0000
   * 2 bytes termination 0x0000
   */
  unsigned char data[2];

  __enter;

  njb3_start_play[4] = (trackid>>24) & 255;
  njb3_start_play[5] = (trackid>>16) & 255;
  njb3_start_play[6] = (trackid>>8) & 255;
  njb3_start_play[7] = trackid & 255;

  if (send_njb3_command (njb, njb3_start_play, 0x10) == -1) {
    __leave;
    return -1;
  }

  if ( (bread= usb_pipe_read(njb, data, 2)) == -1 ) {
    NJB_ERROR(EO_USBBLK);
    __leave;
    return -1;
  }  
  else if (bread < 2) {
    NJB_ERROR (EO_RDSHORT);
    __leave;
    return -1;
  }

  /* Update njb3_current_playing_track. */
  /* The play_track call always puts the track on the front of the playlist. */
  njb->current_playing_track = 0;

  __leave;
  return 0;
}

int njb3_seek_track (njb_t * njb, u_int32_t pos)
{
  __dsub= "njb3_seek_track";

  ssize_t bread;
  u_int16_t result;
  unsigned char njb3_seek[0x0e] = {
    0x00, 0x07, 0x00, 0x01, 0x00, 0x06, 0x01, 0x0c,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00
  };
  /* Structure: 14 bytes
   * 2 bytes unknown 0x0007
   * 2 bytes unknown 0x0001
   * 2 bytes length of following record 0x0006
   * 2 bytes metadata frame name 0x010c
   * 4 bytes track position to seek to (in ms)
   * 2 bytes termination 0x0000
   */

  __enter;

  njb3_seek [8] = (pos >> 24) & 255;
  njb3_seek [9] = (pos >> 16) & 255;
  njb3_seek [10] = (pos >> 8) & 255;
  njb3_seek [11] = pos & 255;

  if (send_njb3_command (njb, njb3_seek, 0x0e) == -1) {
    __leave;
    return -1;
  }

  if ( (bread= usb_pipe_read(njb, &result, 2)) == -1 ) {
    NJB_ERROR(EO_USBBLK);
    __leave;
    return -1;
  }  
  else if (bread < 2) {
    NJB_ERROR (EO_RDSHORT);
    __leave;
    return -1;
  }

  __leave;
  return result;
}

int njb3_ctrl_playing(njb_t * njb, int cmd)
{
  __dsub= "njb3_ctrl_playing";

  ssize_t bread;
  unsigned char njb3_request_playing[0x0c] = {
    0x00, 0x08, 0x00, 0x01, 0xFF, 0xFE, 0x00, 0x02,
    0x01, 0x0b, 0x00, 0x00
  };
  /* Structure: 14 bytes
   * 2 bytes unknown 0x0008 (reading operation ?)
   * 2 bytes unknown 0x0001
   * 2 bytes unknown 0xFFFE
   * 2 bytes request code length 0x0002
   * 2 bytes metadata frame name 0x010b
   * 2 bytes termination 0x0000
   */
  unsigned char njb3_pause[0x0c] = {
    0x00, 0x07, 0x00, 0x01, 0x00, 0x04, 0x01, 0x0b,
    0x0b, 0x00, 0x00, 0x00
  };
  unsigned char play_status[0x0a];
  u_int16_t result;

  __enter;

  if (send_njb3_command (njb, njb3_request_playing, 0x0c) == -1) {
    __leave;
    return -1;
  }

  if ( (bread= usb_pipe_read(njb, play_status, 0x0a)) == -1 ) {
    NJB_ERROR(EO_USBBLK);
    __leave;
    return -1;
  }  
  else if (bread < 0x0a) {
    NJB_ERROR (EO_RDSHORT);
    __leave;
    return -1;
  }

  njb3_pause[9] = cmd;

  /* Don't send redundant commands. */
  switch (cmd) {
  case NJB3_START_PLAY:
    if (play_status[7] != NJB3_STOP_PLAY) {
      __leave;
      return 0;
    }
    break;
  case NJB3_STOP_PLAY:
    njb->current_playing_track = 0;
    if (play_status[7] == NJB3_STOP_PLAY) {
      __leave;
      return 0;
    }
    break;
  case NJB3_PAUSE_PLAY:
    if (play_status[7] != NJB3_START_PLAY) {
      __leave;
      return 0;
    }
    break;
  case NJB3_RESUME_PLAY:
    if (play_status[7] != NJB3_PAUSE_PLAY) {
      __leave;
      return 0;
    }
    break;
  }

  if (send_njb3_command (njb, njb3_pause, 0x0c) == -1) {
    __leave;
    return -1;
  }

  if ( (bread= usb_pipe_read(njb, &result, 2)) == -1 ) {
    NJB_ERROR(EO_USBBLK);
    __leave;
    return -1;
  }  
  else if (bread < 2) {
    NJB_ERROR (EO_RDSHORT);
    __leave;
    return -1;
  }  

  __leave;
  return result;
}

int njb3_set_time(njb_t *njb, njb_time_t *time)
{
	__dsub= "njb3_set_time";
	unsigned char *data;
       	//unsigned char njb3_set_time_command[]={0x00,0x07,0x00,0x01,0x00,0x0a,0x01,0x10,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00};
	/* Structure: 18 bytes
       	 * 2 bytes unknown 0x0007 (could mean "write operation")
       	 * 2 bytes unknown 0x0001
       	 * 2 bytes unknown 0x000a
       	 * 2 bytes metadate frame for time 0x0110
	 * 8 bytes time frame
       	 * 2 bytes termination 0x0000
       	 */

	__enter;
	
	data = time_pack3(time);

       	if (send_njb3_command(njb, data, 0x12) == -1){
	        free(data);
       		__leave;
       		return -1;
       	}

	free(data);
	__leave;
	return 0;
}


static void add_song_to_njb(njb_t *njb, songid_t *song)
{
  /* Insert this into the list */
  if (njb->first_songid == NULL) {
    njb->first_songid = song;
    njb->next_songid = song;
  } else {
    njb->next_songid->next = song;
    njb->next_songid = song;
  }
}


/* 
 * This routine not only gets the first track,
 * but makes a list of *ALL* tracks.  next_track_tag is a dummy
 * that follows this list.
 */
songid_t *njb3_get_first_track_tag (njb_t *njb)
{
	__dsub= "njb3_get_first_track_tag";
	unsigned char njb3_get_track_tag_header[]=
				{0x00,0x06,0x00,0x01,0x00,0x00,0x00,0x02,
				0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
				0x00,0x00,0x01,0x00,0xff,0xfe,0x00,0x18,
				0x01,0x04,0x01,0x02,0x01,0x03,0x01,0x01,
				0x00,0x0e,0x00,0x0b,0x00,0x06,0x01,0x07,
				0x01,0x06,0x01,0x05,0x00,0x0d,0x00,0x07,
				0x00,0x00,0x00,0x00};
	/* Structure: 0x30 bytes
       	 * 2 bytes unknown 0x0006 (could mean "complex read operation")
       	 * 2 bytes unknown 0x0001
       	 * 2 bytes unknown 0x0000
       	 * 2 bytes read track tags 0x0002
	 * 8 bytes previous data, initially 0xff * 8
	 * 2 bytes unknown 0x0000
	 * 2 bytes number of tags to request 0x0100 = 256 (I think.)
	 * 2 bytes unknown 0xfffe
	 *
	 * 2 bytes length of metadata wishlist (which follows 0x14 originally)
	 *
	 * The following is then a list of metadata to be returned
	 *
	 * 2 bytes 0x0104 = return Title metadata
	 * 2 bytes 0x0102 = return Artist metadata
	 * 2 bytes 0x0103 = return Genre metadata
	 * 2 bytes 0x0101 = return Album metadata
	 * 2 bytes 0x000e = return File size metadata
	 * 2 bytes 0x000b = return Locked status metadata
	 * 2 bytes 0x0006 = return Codec metadata
	 * 2 bytes 0x0107 = return Year metadata
	 * 2 bytes 0x0106 = return Track number metadata
	 * 2 bytes 0x0105 = return Runlength metadata
	 * 2 bytes 0x000d = return Unknown metadata (not requested by all apps)
	 * 2 bytes 0x0007 = return Filename metadata (not requested by all apps)
	 *
	 * 2 bytes unknown 0x0000
	 * 2 bytes unknown 0x0000
       	 */
	unsigned char *data;
	int keep_reading_chunks=1;
	int i=0,j=0;
	u_int32_t status;
	u_int32_t framelen;
	u_int32_t frameid;
	u_int32_t track_size;
	u_int32_t track_year;
	u_int32_t track_number;
	u_int32_t tracktime;
	int next_chunk = 1;
	int read_again = 0;
	char *tmp;
	songid_t *song = NULL;
	songid_frame_t *frame;
	ssize_t bread;

	__enter;

	/* Could add a routine to destroy any partial songid list */

	/* Allocate a track scanning buffer */
	if((data=(unsigned char *)malloc(NJB3_METADATA_BUFSIZE))==NULL){
		NJB_ERROR(EO_NOMEM);
		__leave;
		return NULL;
	}

	/* Clean from any previous scan */
	njb->first_songid = NULL;
	njb->next_songid = NULL;

	/* Scan the tracks */

	if (send_njb3_command(njb, njb3_get_track_tag_header, 0x34) == -1){
	  __leave;
	  return NULL;
	}
	if ( (bread= usb_pipe_read(njb, data, NJB3_METADATA_BUFSIZE)) == -1 ) {
	  NJB_ERROR(EO_USBBLK);
	  __leave;
	  return NULL;
	} 

	/* Check the status word */
	status = (data[i]<<8) + data[i+1];
	if (status != 0) {
	  __leave;
	  return NULL;
	}
	i = 2;
	while(i<bread && keep_reading_chunks==1){
	  /* Now we recognize the frames */
	  next_chunk = 0;
	  while (!next_chunk) {
	    /* Length of the next frame */
	    framelen = (data[i]<<8) + (data[i+1]) + 2;

	    /* <Reread> if end-of-data:
	     *
	     * Read data from the jukebox and push it into the data buffer,
	     * behind the current frame.
	     */
	    if (i+framelen >= bread) {
	      /* How many bytes to save (the current frame) */
	      int saveme = bread-i;
	      unsigned char *tmpbuffer;

	      /* Move current frame into the first bytes of the buffer.
	       * If the buffer size is very small (if the jukebox
	       * sends small pieces of metadata), this can be an overlapping
	       * operation. Thus we copy to a temporary buffer and back. */
	      tmpbuffer = (unsigned char *) malloc(saveme);
	      memcpy (tmpbuffer, data+i, saveme);
	      memcpy (data, tmpbuffer, saveme);
	      free(tmpbuffer);

	      /* Push new data from jukebox into the buffer right behind
	       * the current frame. */
	      if ( (bread= usb_pipe_read(njb, data+saveme, NJB3_METADATA_BUFSIZE-saveme)) == -1 ) {
		NJB_ERROR(EO_USBBLK);
		__leave;
		return NULL;
	      }
	      /* bread is set to the whole buffer size, that means it must be
	       * increased by the size of the current frame. */
	      bread += saveme;
	      /*
	      data_dump_ascii (stderr, data, bread, 16);
	      */
	      /* Reset the index. */
	      i = 0;
	    } /* </Reread> Continue as if nothing happens. */

	    /* What are the contents of the next frame */
	    frameid = (data[i+2]<< 8) + data[i+3];
	    
	    /* Zero length frame and framid zero means "read again" */
	    if ((framelen == 2) && (frameid == 0)) {
	      next_chunk = 1;
	      framelen = 4;
	      add_song_to_njb(njb, song);
	      song = NULL;
	    }
	    /* A zero length terminates the tag */
	    else if (framelen == 2) {
	      add_song_to_njb(njb, song);
	      /*
		songid_dump(song, stdout);
		printf("\n-------------\n");
	      */
	      song = NULL;
	    }
	    /* Recognize track ID */
	    else if (frameid == NJB3_POSTID_FRAME_ID) {
	      /* We detected the start of a track tag */
	      /* Then create a new song ID */
	      song = songid_new();
	      if (song == NULL) {
		__leave;
		return NULL;
	      }
	      song->trid = (data[i+4]<<24) + (data[i+5]<<16) + (data[i+6]<<8) + (data[i+7]);
	    }
	    /* 0x01 0x04 = Track name, create a frame with the track name */
	    else if (frameid == NJB3_TITLE_FRAME_ID) {
	      tmp = ucs2tostr(&data[i+4]); // Check for out of mem
	      frame = songid_frame_new(strlen(FR_TITLE)+1, FR_TITLE,
				       ID_DATA_ASCII, strlen(tmp)+1, tmp); // Check for out of mem
	      free(tmp);
	      songid_addframe(song, frame);
	    }
	    /* Create a frame with the artist name */
	      else if (frameid == NJB3_ARTIST_FRAME_ID) {
		tmp = ucs2tostr(&data[i+4]); // Check for out of mem
		frame = songid_frame_new(strlen(FR_ARTIST)+1, FR_ARTIST,
					 ID_DATA_ASCII, strlen(tmp)+1, tmp); // Check for out of mem
		free(tmp);
		songid_addframe(song, frame);
	      }
	    /* Create a frame with the genre name */
	    else if (frameid == NJB3_GENRE_FRAME_ID) {
	      tmp = ucs2tostr(&data[i+4]); // Check for out of mem
	      frame = songid_frame_new(strlen(FR_GENRE)+1, FR_GENRE,
				       ID_DATA_ASCII, strlen(tmp)+1, tmp); // Check for out of mem
	      free(tmp);
	      songid_addframe(song, frame);
	    }
	    /* Create a frame with the album name */
	    else if (frameid ==  NJB3_ALBUM_FRAME_ID) {
	      tmp = ucs2tostr(&data[i+4]); // Check for out of mem
	      frame = songid_frame_new(strlen(FR_ALBUM)+1, FR_ALBUM,
				       ID_DATA_ASCII, strlen(tmp)+1, tmp); // Check for out of mem
	      free(tmp);
	      songid_addframe(song, frame);
	    }
	    /* Create a frame with the track size */
	    else if (frameid ==  NJB3_FILESIZE_FRAME_ID) {
	      track_size = (data[i+4] << 24) + (data[i+5] << 16) + (data[i+6] << 8) + (data[i+7]);
	      /*
		Old erroneous code just for reference (delete after 2003-05-01)
		tmp = int32tostring(track_size);
		frame = songid_frame_new(strlen(FR_SIZE)+1, FR_SIZE,
		ID_DATA_ASCII, strlen(tmp)+1, tmp); // Check for out of mem
		free(tmp);
	      */
	      frame = songid_frame_new(strlen(FR_SIZE)+1, FR_SIZE,
				       ID_DATA_BIN, 4, &track_size); // Check for out of mem
	      songid_addframe(song, frame);
	    }
	    else if (frameid ==  NJB3_LOCKED_FRAME_ID) {
	      /*
	       * It seems the value 0x0100 means "locked", 0x0000 means "open"
	       * printf("Locked Track? Value %02x%02x\n", data[i+4], data[i+5]);
	       *
	       */
	      if (data[i+4] == 0x01 && data[i+5] == 0x00) {
		frame = songid_frame_new(strlen(FR_PROTECTED)+1, FR_PROTECTED,
					 ID_DATA_ASCII, 2, "1"); // Check for out of mem
		songid_addframe(song, frame);
	      }
	    }
	    /* Create a frame with codec type */
	    else if (frameid ==  NJB3_CODEC_FRAME_ID) {
	      u_int32_t codec;
	      
	      codec = (data[i+4] << 8) + data [i+5];
	      if (codec == NJB3_CODEC_MP3_ID || codec == NJB3_CODEC_PRELOADED_MP3_ID) {
		frame = songid_frame_new(strlen(FR_CODEC)+1, FR_CODEC,
					 ID_DATA_ASCII, strlen(NJB_CODEC_MP3)+1, NJB_CODEC_MP3);
		// Check for out of mem
		songid_addframe(song, frame);
	      } else if (codec == NJB3_CODEC_WAV_ID) {
		frame = songid_frame_new(strlen(FR_CODEC)+1, FR_CODEC,
					 ID_DATA_ASCII, strlen(NJB_CODEC_WAV)+1, NJB_CODEC_WAV);
		// Check for out of mem
		songid_addframe(song, frame);
	      } else if (codec == NJB3_CODEC_WMA_ID) {
		frame = songid_frame_new(strlen(FR_CODEC)+1, FR_CODEC,
					 ID_DATA_ASCII, strlen(NJB_CODEC_WMA)+1, NJB_CODEC_WMA);
		// Check for out of mem
		songid_addframe(song, frame);
	      } else {
		/* Ogg Vorbis? :-) */
		/*
		  printf("Unknown codec ID %02x%02x\n", data[i+4], data[i+5]);
		*/
	      }
	    }
	    /* Create a frame with the track year */
	    else if (frameid == NJB3_YEAR_FRAME_ID) {
	      track_year = (data[i+4] << 8) + (data[i+5]);
	      tmp = int32tostring(track_year);
	      frame = songid_frame_new(strlen(FR_YEAR)+1, FR_YEAR,
				       ID_DATA_ASCII, strlen(tmp)+1, tmp); // Check for out of mem
	      free(tmp);
	      songid_addframe(song, frame);
	    }
	    /* Create a frame with the track number */
	    else if (frameid == NJB3_TRACKNO_FRAME_ID) {
	      track_number = (data[i+4] << 8) + (data[i+5]);
	      tmp = int32tostring(track_number);
	      frame = songid_frame_new(strlen(FR_TRACK)+1, FR_TRACK,
				       ID_DATA_ASCII, strlen(tmp)+1, tmp); // Check for out of mem
	      free(tmp);
	      songid_addframe(song, frame);
	    }
	    /* Create a frame with the track runlengh (time) */
	    else if (frameid == NJB3_LENGTH_FRAME_ID) {
	      tracktime=(data[i+4] << 8) + (data[i+5]);
	      frame = songid_frame_new(strlen(FR_LENGTH)+1, FR_LENGTH,
				       ID_DATA_BIN, 4, &tracktime); // +2? +4 is OK
	      songid_addframe(song, frame);
	    }
	    /* Create a frame with the original filename */
	    else if (frameid == NJB3_FNAME_FRAME_ID) {
	      tmp = ucs2tostr(&data[i+4]); // Check for out of mem
	      frame = songid_frame_new(strlen(FR_FNAME)+1, FR_FNAME,
				       ID_DATA_ASCII, strlen(tmp)+1, tmp); // Check for out of mem
	      free(tmp);
	      songid_addframe(song, frame);
	    }
	    /* Ignore data about the track directory */
	    else if (frameid == NJB3_DIR_FRAME_ID) {
	      tmp = ucs2tostr(&data[i+4]); // Check for out of mem
	      /* Currently we just ignore this string, could define a new FR_DIR or so */
	      free(tmp);
	    }
	    /* Unknown frame */
	    else {
	      /* 
		 printf("Unknown frame %02x%02x, length %02x%02x bytes\n",
		 data[i+2],data[i+3],data[i],data[i+1]);
	      */
	    }
	    /* Increase the counter */
	    i += framelen;
	  }
	  read_again = (data[i+8]<<8) + data[i+9];
	  /* If this is zero, read again */
	  if (read_again == 0) {
	    for(j = i; j < i+8; j++) {
	      njb3_get_track_tag_header[j-i+8]=data[j];
	    }
	    if (send_njb3_command(njb, njb3_get_track_tag_header, 0x34) == -1){
	      __leave;
	      return NULL;
	    }
	    if ( (bread= usb_pipe_read(njb, data, NJB3_METADATA_BUFSIZE)) == -1 ) {
	      NJB_ERROR(EO_USBBLK);
	      __leave;
	      return NULL;
	    }

	    /* Check the status word */
	    status = (data[0]<<8) + data[1];

	    /* NJB2 sometimes(?) sends a 0x000e 0x0000 0x01f4 0x001b 0x0001 0x0000 0x0001
	     * on end-of-data.
	     */
	    if (status == 0x000e)
	      keep_reading_chunks = 0;
	    else if (status != 0) {
	      __leave;
	      return NULL;
	    }
	    i=2;
	  } else {
	    /* End this lollapalooza */
	    keep_reading_chunks = 0;
	  }
	}
	free(data);
	
	__leave;

	if (njb->first_songid != NULL) {
		njb->next_songid = njb->first_songid->next;
	}
	return njb->first_songid;
}


songid_t *njb3_get_next_track_tag (njb_t *njb)
{
  /* Returns a songID from the list */
  songid_t *tmp;

  if (njb->next_songid == NULL) {
    njb->first_songid = NULL;
    return NULL;
  }
  tmp = njb->next_songid;
  njb->next_songid = tmp->next;
  return tmp;
}

static int populate_playlist_tag(njb_t *njb, playlist_t *pl)
{
	__dsub= "populate_playlist_tag";
	unsigned char njb3_get_tracks_for_pl[]={0x01,0x08,0x00,0x01,0x00,0x00,0x00,0x00,
						0x00,0x00,0x01,0x00,0xff,0xfe,0x00,0x02,
						0x00,0x0c,0x00,0x00};
	/* Structure: 0x14 bytes
       	 * 2 bytes unknown 0x0108
       	 * 2 bytes unknown 0x0001
       	 * 4 bytes entity ID
       	 * 2 bytes unknown 0x0000
       	 * 2 bytes unknown 0x0100
       	 * 2 bytes unknown 0xfffe
       	 * 2 bytes request code length 0x0002
       	 * 2 bytes metadata frame name 0x000c (POSTID)
       	 * 2 bytes termination 0x0000
	 */
	unsigned char *data;
	int keep_reading_chunks=1;
	int next_chunk = 1;
	int read_again = 0;
	int i=0;
	u_int32_t status;
	u_int32_t framelen;
	u_int32_t frameid;
	u_int32_t trackid = 0;
	ssize_t bread;
	playlist_track_t *track;

	__enter;

	from_32bit_to_njb3_bytes(pl->plid, &njb3_get_tracks_for_pl[4]);

	/* Allocate a playlist tracklist buffer */
	if((data=(unsigned char *)malloc(NJB3_METADATA_BUFSIZE))==NULL){
		NJB_ERROR(EO_NOMEM);
		__leave;
		return -1;
	}

	/* Scan the playlists */
	if (send_njb3_command(njb, njb3_get_tracks_for_pl, 0x14) == -1){
	  __leave;
	  return -1;
	}
	if ( (bread= usb_pipe_read(njb, data, NJB3_METADATA_BUFSIZE)) == -1 ) {
	  NJB_ERROR(EO_USBBLK);
	  __leave;
	  return -1;
	}

	/* Check the status word */
	status = njb3_bytes_to_16bit(&data[i]);
	/*
	  I don't know what this word means, it's usually 0x0e
	  doesn't seem to be status, not the number of tracks
	  in the list either.
	  Could be:
	  0x000e = Playlist tracks follow
	  0x0000 = Empty playlist
	*/

	/* This makes us leave the party if there are no songs in this
	 * playlist
	 */
	if (status == 0x0000) {
	  __leave;
	  return 0;
	}
	i = 2;
	while(i<bread && keep_reading_chunks==1){
	  /* Now we recognize the frames */
	  next_chunk = 0;
	  while (!next_chunk) {
	    /* Length of the next frame */
	    framelen = (data[i]<<8) + (data[i+1]) + 2;

	    if (i+framelen >= bread) {
	      /* How many bytes to save (the current frame) */
	      int saveme = bread-i;
	      unsigned char *tmpbuffer;

	      /* Move current frame into the first bytes of the buffer.
	       * If the buffer size is very small (if the jukebox
	       * sends small pieces of metadata), this can be an overlapping
	       * operation. Thus we copy to a temporary buffer and back. */
	      tmpbuffer = (unsigned char *) malloc(saveme);
	      memcpy (tmpbuffer, data+i, saveme);
	      memcpy (data, tmpbuffer, saveme);
	      free(tmpbuffer);

	      /* Push new data from jukebox into the buffer right behind
	       * the current frame. */
	      if ( (bread= usb_pipe_read(njb, data+saveme, NJB3_METADATA_BUFSIZE-saveme)) == -1 ) {
		NJB_ERROR(EO_USBBLK);
		__leave;
		return -1;
	      }
	      /* bread is set to the whole buffer size, that means it must be
	       * increased by the size of the current frame. */
	      bread += saveme;

	      /* Reset the index. */
	      i = 0;
	    } /* </Reread> Continue as if nothing happens. */


	    /* What are the contents of the next frame */
	    frameid = (data[i+2]<< 8) + data[i+3];

	    /* Zero length frame and framid zero means "read again" */
	    if ((framelen == 2) && (frameid == 0x0000)) {
	      track = playlist_track_new (trackid);
	      playlist_addtrack (pl, track, NJB_PL_END);
	      next_chunk = 1;
	      if (data[i+4] == 0x00 && data[i+5] == 0x01)
		read_again = 0;
	      else
		read_again++;
	    }
	    /* A zero length terminates the tag */
	    else if (framelen == 2) {
	      track = playlist_track_new (trackid);
	      playlist_addtrack (pl, track, NJB_PL_END);
	    }
	    /* Recognize track ID */
	    else if (frameid == NJB3_POSTID_FRAME_ID) {
	      /* We detected the start of a track tag */
	      /* Then create a new playlist ID */
	      trackid = (data[i+4]<<24) + (data[i+5]<<16) + (data[i+6]<<8) + (data[i+7]);
	    }
	    /* Unknown frame */
	    else {
	      /* 
	       printf("Unknown playlist frame %02x%02x, length %02x%02x bytes\n",
	       data[i+2],data[i+3],data[i],data[i+1]);
	      */
	    }
	    /* Increase the counter */
	    i += framelen;
	  }
	  if (read_again > 0) {
	    njb3_get_tracks_for_pl[8] = read_again & 0xFF;
	    if (send_njb3_command (njb, njb3_get_tracks_for_pl, 0x14) == -1) {
	      __leave;
	      return -1;
	    }
	    if ((bread = usb_pipe_read (njb, data, NJB3_METADATA_BUFSIZE)) == -1) {
	      NJB_ERROR(EO_USBBLK);
	      __leave;
	      return -1;
	    }

	    /* Check the status word */
	    status = (data[0]<<8) + data[1];
	    i=2;
	  }
	  else {
	    keep_reading_chunks = 0;
	  }
	}
	free(data);

	__leave;
	return 0;
}


static int add_pl_to_njb(njb_t *njb, playlist_t *pl)
{
  /* Populate the playlist with tracks */
  if (populate_playlist_tag(njb, pl) == -1) {
    return -1;
  }
  pl->_state = NJB_PL_UNCHANGED;
  /* Insert this into the list */
  if (njb->first_plid == NULL) {
    njb->first_plid = pl;
    njb->next_plid = pl;
  } else {
    njb->next_plid->nextpl = pl;
    njb->next_plid = pl;
  }
  return 0;
}

playlist_t *njb3_get_first_playlist_tag (njb_t *njb)
{
	__dsub= "njb3_get_first_playlist_tag";
	unsigned char njb3_get_playlists[]={0x00,0x06,0x00,0x01,0x00,0x00,0x00,0x01,
					    0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
					    0x00,0x00,0x01,0x00,0xff,0xfe,0x00,0x02,
					    0x01,0x0f,0x00,0x00,0x00,0x00};
	/* Structure: 0x1E bytes
       	 * 2 bytes unknown 0x0006 (could mean "complex read operation")
       	 * 2 bytes unknown 0x0001
       	 * 2 bytes unknown 0x0000
       	 * 2 bytes read playlist tags 0x0001
	 * 8 bytes previous data, initially 0xff * 8
	 * 2 bytes unknown 0x0000
	 * 2 bytes maximum number of lists to request 0x0100 = 256 (I think.)
	 * 2 bytes unknown 0xfffe
	 *
	 * 2 bytes length of playlist metadata wishlist
	 *
	 * 2 bytes 0x010f = return Playlist names
	 *
	 * 2 bytes unknown 0x0000
	 * 2 bytes unknown 0x0000
       	 */
	unsigned char *data;
	int keep_reading_chunks=1;
	int i=0,j=0;
	u_int16_t status;
	u_int16_t framelen;
	u_int16_t frameid;
	int next_chunk = 1;
	u_int16_t read_again = 0;
	char *tmp;
	ssize_t bread;
	playlist_t *pl = NULL;

	__enter;

	/* Could add a routine to destroy any partial songid list */

	/* Allocate a playlist scanning buffer */
	if((data=(unsigned char *)malloc(NJB3_METADATA_BUFSIZE))==NULL){
		NJB_ERROR(EO_NOMEM);
		__leave;
		return NULL;
	}

	/* Clean from any previous scan */
	njb->first_plid = NULL;
	njb->next_plid = NULL;

	/* Scan the playlists */
	if (send_njb3_command(njb, njb3_get_playlists, 0x1e) == -1){
	  __leave;
	  return NULL;
	}
	if ( (bread= usb_pipe_read(njb, data, NJB3_METADATA_BUFSIZE)) == -1 ) {
	  NJB_ERROR(EO_USBBLK);
	  __leave;
	  return NULL;
	}
	/* Check the status word */
	status = njb3_bytes_to_16bit(&data[i]);
	if (status != 0) {
	  __leave;
	  return NULL;
	}

	i += 2;
	while(i<bread && keep_reading_chunks==1){
	  /* Now we recognize the frames */
	  next_chunk = 0;
	  while (!next_chunk) {
	    /* Length of the next frame */
	    framelen = njb3_bytes_to_16bit(&data[i]) + 2;

	    if (i+framelen >= bread) {
	      /* How many bytes to save (the current frame) */
	      int saveme = bread-i;

	      /* Move current frame into the first bytes of the buffer.
	       * CAUTION: If the buffer size is very small (if the jukebox
	       * sends small pieces of metadata), this can be an overlapping
	       * operation. */
	      memcpy (data, data+i, saveme);

	      /* Push new data from jukebox into the buffer right behind
	       * the current frame. */
	      if ( (bread= usb_pipe_read(njb, data+saveme, NJB3_METADATA_BUFSIZE-saveme)) == -1 ) {
		NJB_ERROR(EO_USBBLK);
		__leave;
		return NULL;
	      }
	      /* bread is set to the whole buffer size, that means it must be
	       * increased by the size of the current frame. */
	      bread += saveme;
	      /*
	      data_dump_ascii (stderr, data, bread, 16);
	      */
	      /* Reset the index. */
	      i = 0;
	    } /* </Reread> Continue as if nothing happens. */

	    /* What are the contents of the next frame */
	    frameid = njb3_bytes_to_16bit(&data[i+2]);

	    /* Zero length frame and framid zero means "read again" */
	    if ((framelen == 2) && (frameid == 0)) {
	      next_chunk = 1;
	      framelen = 4;
	      if (add_pl_to_njb(njb, pl) == -1) {
		__leave;
		return NULL;
	      }
	      pl = NULL;
	    }
	    /* A zero length terminates the tag */
	    else if (framelen == 2) {
	      if (add_pl_to_njb(njb, pl) == -1) {
		__leave;
		return NULL;
	      }
	      pl = NULL;
	    }
	    /* Recognize track ID */
	    else if (frameid == NJB3_POSTID_FRAME_ID) {
	      /* We detected the start of a track tag */
	      /* Then create a new playlist ID */
	      pl = playlist_new();
	      if (pl == NULL) {
		__leave;
		return NULL;
	      }
	      pl->plid = njb3_bytes_to_32bit(&data[i+4]);
	      pl->ntracks = 0;
	    }
	    /* Create a frame with the playlist name */
	    else if (frameid == NJB3_PLNAME_FRAME_ID) {
	      tmp = ucs2tostr(&data[i+4]);
	      pl->name = tmp;
	    }
	    /* Unknown frame */
	    else {
	      /*
	      printf("Unknown playlist frame %02x%02x, length %02x%02x bytes\n",
		     data[i+2],data[i+3],data[i],data[i+1]);
	      */
	    }
	    /* Increase the counter */
	    i += framelen;
	  }
	  read_again = njb3_bytes_to_16bit(&data[i+8]);
	  /* If this is zero, read again */
	  if (read_again == 0) {
	    for(j = i; j < i+8; j++) {
	      njb3_get_playlists[j-i+8]=data[j];
	    }
	    if (send_njb3_command(njb, njb3_get_playlists, 0x1e) == -1){
	      __leave;
	      return NULL;
	    }
	    if ( (bread= usb_pipe_read(njb, data, NJB3_METADATA_BUFSIZE)) == -1 ) {
	      NJB_ERROR(EO_USBBLK);
	      __leave;
	      return NULL;
	    }
	    /* Check the status word */
	    status = njb3_bytes_to_16bit(&data[0]);
	    if (status != 0) {
	      __leave;
	      return NULL;
	    }
	    i=2;
	  } else {
	    /* End this lollapalooza */
	    keep_reading_chunks = 0;
	  }
	}
	free(data);
	
	__leave;

	if (njb->first_plid != NULL) {
		njb->next_plid = njb->first_plid->nextpl;
	}
	return njb->first_plid;
}

playlist_t *njb3_get_next_playlist_tag (njb_t *njb)
{
  /* Returns a playlistID from the list */
  playlist_t *tmp;

  if (njb->next_plid == NULL) {
    njb->first_plid = NULL;
    return NULL;
  }
  tmp = njb->next_plid;
  njb->next_plid = tmp->nextpl;
  return tmp;
}

static int add_df_to_njb(njb_t *njb, datafile_t *df)
{
  /* Insert this into the list */
  if (njb->first_dfid == NULL) {
    njb->first_dfid = df;
    njb->next_dfid = df;
  } else {
    njb->next_dfid->nextdf = df;
    njb->next_dfid = df;
  }
  return 0;
}

datafile_t *njb3_get_first_datafile_tag (njb_t *njb) {
	__dsub= "njb3_get_first_datafile_tag";
	unsigned char njb3_get_datafiles[]={0x00,0x06,0x00,0x01,0x00,0x00,0x00,0x00,
					    0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
					    0x00,0x00,0x02,0x00,0xff,0xfe,0x00,0x0a,
					    0x00,0x07,0x00,0x0e,0x00,0x0d,0x00,0x16,
					    0x00,0x06,0x00,0x00,0x00,0x00};
	/* Structure: 0x26 bytes
       	 * 2 bytes unknown 0x0006 (could mean "complex read operation")
       	 * 2 bytes unknown 0x0001
       	 * 2 bytes unknown 0x0000
       	 * 2 bytes read datafile tags 0x0000
	 * 8 bytes previous data, initially 0xff * 8
	 * 2 bytes unknown 0x0000
	 * 2 bytes maximum number of lists to request 0x0200 = 256 (I think.)
	 * 2 bytes unknown 0xfffe
	 *
	 * 2 bytes length of datafile metadata wishlist
	 *
	 * 2 bytes 0x0007 = return file names
	 * 2 bytes 0x000e = return file size
	 * 2 bytes 0x000d = return directory name
	 * 2 bytes 0x0016 = return ... something (unknown)
	 * 2 bytes 0x0006 = return locked status
	 *
	 * 2 bytes unknown 0x0000
	 * 2 bytes unknown 0x0000
       	 */
	unsigned char *data;
	int keep_reading_chunks=1;
	int i=0,j=0;
	u_int16_t status;
	u_int16_t framelen;
	u_int16_t frameid;
	int next_chunk = 1;
	int read_again = 0;
	char *tmp;
	ssize_t bread;
	datafile_t *df = NULL;

	__enter;

	/* Could add a routine to destroy any partial songid list */

	/* Allocate a datafile scanning buffer */
	if((data=(unsigned char *)malloc(NJB3_METADATA_BUFSIZE))==NULL){
		NJB_ERROR(EO_NOMEM);
		__leave;
		return NULL;
	}

	/* Clean from any previous scan */
	njb->first_dfid = NULL;
	njb->next_dfid = NULL;

	/* Scan the datafiles */

	if (send_njb3_command(njb, njb3_get_datafiles, 0x26) == -1){
	  __leave;
	  return NULL;
	}
	if ( (bread= usb_pipe_read(njb, data, NJB3_METADATA_BUFSIZE)) == -1 ) {
	  NJB_ERROR(EO_USBBLK);
	  __leave;
	  return NULL;
	}

	/* Check the status word */
	status = njb3_bytes_to_16bit(&data[i]);
	if (status != 0) {
	  __leave;
	  return NULL;
	}
	i = 2;
	while(i<bread && keep_reading_chunks==1){
	  /* Now we recognize the frames */
	  next_chunk = 0;
	  while (!next_chunk) {
	    /* Length of the next frame */
	    framelen = njb3_bytes_to_16bit(&data[i]) + 2;
	    /* What are the contents of the next frame */
	    frameid = njb3_bytes_to_16bit(&data[i+2]);
	    
	    /* Zero length frame and framid zero means "read again" */
	    if ((framelen == 2) && (frameid == 0)) {
	      next_chunk = 1;
	      framelen = 4;
	      if (add_df_to_njb(njb, df) == -1) {
		__leave;
		return NULL;
	      }
	      df = NULL;
	    }
	    /* A zero length terminates the tag */
	    else if (framelen == 2) {
	      if (add_df_to_njb(njb, df) == -1) {
		__leave;
		return NULL;
	      }
	      df = NULL;
	    }
	    /* Recognize track ID */
	    else if (frameid == NJB3_POSTID_FRAME_ID) {
	      /* We detected the start of a track tag */
	      /* Then create a new datafile ID */
	      df = datafile_new();
	      if (df == NULL) {
		__leave;
		return NULL;
	      }
	      df->dfid = njb3_bytes_to_32bit(&data[i+4]);
	    }
	    /* Create a frame with the datafile name */
	    else if (frameid == NJB3_FNAME_FRAME_ID) {
	      tmp = ucs2tostr(&data[i+4]); // Check for out of mem
	      df->filename = tmp;
	    }
	    else if(frameid == NJB3_FILESIZE_FRAME_ID) {
	      df->lsdw = njb3_bytes_to_32bit(&data[i+4]);
	      df->msdw = 0;
	    }
	    /* Ignore data about the track directory */
	    else if (frameid == NJB3_DIR_FRAME_ID) {
	      tmp = ucs2tostr(&data[i+4]); // Check for out of mem
	      /* Currently we just ignore this string, could define a new FR_DIR or so */
	      free(tmp);
	    }
	    else if (frameid ==  NJB3_LOCKED_FRAME_ID) {
	      /*
	       * It seems the value 0x0100 means "locked", 0x0000 means "open"
	       * printf("Locked File? Value %02x%02x\n", data[i+4], data[i+5]);
	       * don't know how to handle protected files yet, do they actually
	       * occur?
	       */
	    }
	    /* Unknown or ignored frame */
	    else {
	      /* 
		 printf("Unknown datafile frame %02x%02x, length %02x%02x bytes\n",
		 data[i+2],data[i+3],data[i],data[i+1]);
	      */
	    }
	    /* Increase the counter */
	    i += framelen;
	  }
	  read_again = njb3_bytes_to_16bit(&data[i+8]);
	  /* If this is zero, read again */
	  if (read_again == 0) {
	    for(j = i; j < i+8; j++) {
	      njb3_get_datafiles[j-i+8]=data[j];
	    }
	    if (send_njb3_command(njb, njb3_get_datafiles, 0x26) == -1){
	      __leave;
	      return NULL;
	    }
	    if ( (bread= usb_pipe_read(njb, data, NJB3_METADATA_BUFSIZE)) == -1 ) {
	      NJB_ERROR(EO_USBBLK);
	      __leave;
	      return NULL;
	    }
	    /* Check the status word */
	    status = njb3_bytes_to_16bit(&data[0]);
	    if (status != 0) {
	      __leave;
	      return NULL;
	    }
	    i=2;
	  } else {
	    /* End this lollapalooza */
	    keep_reading_chunks = 0;
	  }
	}
	free(data);
	
	__leave;

	if (njb->first_dfid != NULL) {
	  njb->next_dfid = njb->first_dfid->nextdf;
	}
	return njb->first_dfid;
}

datafile_t *njb3_get_next_datafile_tag (njb_t *njb) {
  /* Returns a datafile tag from the list */
  datafile_t *tmp;

  if (njb->next_dfid == NULL) {
    njb->first_dfid = NULL;
    return NULL;
  }
  tmp = njb->next_dfid;
  njb->next_dfid = tmp->nextdf;
  return tmp;

}

/* Returns file size or -1 if failed.
 * FIXME: As the file size can probably exceed the range of an int,
 *        a u_int32_t should probably be returned from this function
 *        instead. Also fix the verification below.
 */
int njb3_request_file(njb_t *njb, u_int32_t fileid, u_int32_t size)
{
	__dsub= "njb3_request_file";
	unsigned char njb3_request_file[]={0x00,0x02,0x00,0x01,0x00,0x00,0x00,0x00,0x00,0x00,
					   0x00,0x00,0x00,0x10,0x00,0x00};
	/* Structure: 16 bytes
       	 * 2 bytes unknown 0x0002 (could mean "big read operation")
       	 * 2 bytes unknown 0x0001
       	 * 4 bytes file number
       	 * 4 bytes offset 0x0000 0x0000  -> start at 0x0000 0x0000 of the file
	 * 4 bytes transfer chunk maxlength
	 *
	 * Originally maxlength was set to 0x0010 0x0000 but setting it to
	 * big values swallows a lot more files :-) How is maxlength 
	 * supposed to work anyway? Examining transfers of files of
	 * different sizes might yield the answer. The algorithm below
	 * is just a crude guess (but works).
       	 */
	unsigned char status_data[]={0x00,0x00,0x00,0x00,0x00,0x00};
	u_int32_t maxsize = 0x100000;
	u_int32_t filesize;
	int status;
	ssize_t bread;

	__enter;

	/* Set a max transfer size that is > than the reported file size,
	 * I _think_ this might be how other software does the trick, but
	 * I'm not certain. Could be that several njb3_request_file-
	 * commands are sent with different start offsets inside the
	 * file to subdivide it into handlabe transfers. Test this! 
	 * Compare to the 0-length chunk request sent to verify the
	 * file on completion. (Below.) */
	while (maxsize < size)
	  maxsize = (maxsize << 1);
	from_32bit_to_njb3_bytes(maxsize, &njb3_request_file[12]);

	/* Then the actual request command */
	from_32bit_to_njb3_bytes(fileid, &njb3_request_file[4]);

       	if (send_njb3_command(njb, njb3_request_file, 0x10) == -1){
       		__leave;
       		return -1;
       	}
       	if ( (bread= usb_pipe_read(njb, status_data, 6)) == -1 ) {
       		NJB_ERROR(EO_USBBLK);
       		__leave;
       		return -1;
       	} else if ( bread < 6 ) {
       		NJB_ERROR(EO_RDSHORT);
       		__leave;
       		return -1;
       	}
	status = njb3_bytes_to_16bit(&status_data[0]);
	/*
	 * status 0000 = bytes follow
	 * status 0002 = bad request
	 * status 000c = file protected
	 * status 000e = end of file
	 */
	if (status != 0){
	  __leave;
	  return -1;
	}
	filesize = njb3_bytes_to_32bit(&status_data[2]);

	__leave;
	return filesize;
}

int njb3_get_file_chunk(njb_t *njb, int fd, u_int32_t size)
{
	__dsub= "njb3_get_file_chunk";
	unsigned char *data;
	ssize_t bread;
	ssize_t bwritten;

	__enter;

	data = (unsigned char *) malloc(size);
	if (data == NULL) {
		NJB_ERROR(EO_NOMEM);
		__leave;
		return -1;
	}
       	if ( (bread= usb_pipe_read(njb, data, size)) == -1 ) {
		free(data);
       		NJB_ERROR(EO_USBBLK);
       		__leave;
       		return -1;
       	} else if ( bread < size ) {
		free(data);
       		NJB_ERROR(EO_RDSHORT);
       		__leave;
       		return -1;
       	}

	if ( (bwritten= Write(fd, data, size, 0)) == -1 ) {
		free(data);
       		NJB_RERROR("Write", -1);
       		NJB_ERROR(EO_WRFILE);
       		__leave;
       		return -1;
       	} else if ( bwritten != size ) {
		free(data);
       		NJB_ERROR(EO_WRFILE);
       		__leave;
       		return -1;
       	}
	
	free(data);

	__leave;
	return 0;
}

/*
 * FIXME: the verification command *should* be sent. It is the same as the
 *        request command for a size 0 length. Might need some tweaking of
 *        the above and calling functions in procedure.c.
 */
int njb3_get_file_confirmation(njb_t *njb)
{
	__dsub= "njb3_get_file_confirmation";
	/*
	  unsigned char njb3_verify_file[]={0x00,0x02,0x00,0x01,0x00,0x00,0x00,0x00,
					   0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00};
	*/

	/* This command is supposed to be sent after the confirmation byte was recieved.
	 * We don't use it right now becuse we can't compose it. Neither does it
	 * cause any problems not to send it.
	 * Structure: 16 bytes
       	 * 2 bytes unknown 0x0002
       	 * 2 bytes unknown 0x0001
	 * 4 bytes track id
	 * 4 bytes number of bytes transfered + 1 (c801 for an c800 byte file e.g.)
	 *         this is exactly the same number that is returned as file length when
	 *         calling njb3_request_file().
	 * 4 bytes transfer chunk length - must probably be the same as calculated
	 *         inside njb3_request_file().
	 *
	 * After the verification the host reads a 6 bytes status code, which is also
	 * unclear. Example: 0x000e 0x0000 0x0000. The four last bytes should probably
	 * be ignored (it's probably bytes left to send) 0x000e is probably some kind
	 * of status code. (0x000e = EOF).
       	 */

	__enter;

	/*
	 * Here, the verification command would be sent if we knew how it
	 * worked. However, the NJBs will accept the file without verification.
	 * If the transmission was erraneous, the file should be deleted and
	 * possibly retried. So this has to be reverse-engineered correctly.
	 * Please implement this if you find out how it works!
	 */

	__leave;
	return 0;
}

u_int32_t njb3_create_file(njb_t *njb, unsigned char *ptag, u_int32_t tagsize, u_int16_t filetype)
{
	__dsub= "njb3_create_file";
       	unsigned char njb3_send_metadata_tag[]={0x00,0x04,0x00,0x01,0x00,0x00,0x00,0x00};
	u_int32_t bread;
	u_int32_t status;
	u_int32_t fileid;
	unsigned char *data;
	unsigned char status_data[]={0x00,0x00,0x00,0x00,0x00,0x00};
	/* Structure: 8 bytes + n tags
       	 * 2 bytes unknown 0x0004
       	 * 2 bytes unknown 0x0001
       	 * 2 bytes unknown 0x0000
       	 * 2 bytes filetype 0x0002 for tracks 0x0000 for datafiles
	 * after this the tags follow
       	 */
	__enter;

	/* Command size plus tag size */
	data= malloc(8+tagsize);
	if (data == NULL) {
		NJB_ERROR(EO_NOMEM);
		__leave;
		return 0;
	}
	memcpy(&data[0], njb3_send_metadata_tag, 8);
	/* Set file type */
	from_16bit_to_njb3_bytes(filetype, &data[6]);
	/* Copy tag to metadata chunk */
	memcpy(&data[8], ptag, tagsize);
       	if (send_njb3_command(njb, data, 8+tagsize) == -1){
		free(data);
       		__leave;
       		return 0;
       	}
       	if ( (bread= usb_pipe_read(njb, status_data, 6)) == -1 ) {
		free(data);
       		NJB_ERROR(EO_USBBLK);
       		__leave;
       		return 0;
       	} else if ( bread < 6 ) {
		free(data);
       		NJB_ERROR(EO_RDSHORT);
       		__leave;
       		return 0;
       	}
	/* The meanings of different values for this status byte is currently unknown */
	status = njb3_bytes_to_16bit(&status_data[0]);
	if (status == 0) {
	  fileid = njb3_bytes_to_32bit(&status_data[2]);
	} else {
	  fileid = 0;
	}
	free(data);

	__leave;
	return fileid;
}

int njb3_send_file_chunk(njb_t *njb, unsigned char *chunk, u_int32_t chunksize, u_int32_t fileid)
{
	__dsub= "njb3_send_file_chunk";
       	unsigned char njb3_send_file_chunk[]={0x00,0x03,0x00,0x01,0x00,0x00,0x00,0x00,
					       0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00};
	u_int32_t bread;
	u_int32_t status;
	u_int32_t storedsize;
	unsigned char status_data[]={0x00,0x00,0x00,0x00,0x00,0x00};
	/* Structure: 8 bytes
       	 * 2 bytes unknown 0x0003
       	 * 2 bytes unknown 0x0001
	 * 4 bytes file ID
       	 * 2 bytes unknown 0x0000
       	 * 2 bytes unknown 0x0000
	 * 4 bytes file chunk size
       	 */
	
	__enter;

	/* Prepare the command */
	njb3_send_file_chunk[4] = (fileid >> 24) & 255;
	njb3_send_file_chunk[5] = (fileid >> 16) & 255;
	njb3_send_file_chunk[6] = (fileid >> 8) & 255;
	njb3_send_file_chunk[7] = fileid & 255;
	njb3_send_file_chunk[12] = (chunksize >> 24) & 255;
	njb3_send_file_chunk[13] = (chunksize >> 16) & 255;
	njb3_send_file_chunk[14] = (chunksize >> 8) & 255;
	njb3_send_file_chunk[15] = chunksize & 255;

       	if (send_njb3_command(njb, njb3_send_file_chunk, 16) == -1){
       		__leave;
       		return -1;
       	}
       	if (send_njb3_command(njb, chunk, chunksize) == -1){
       		__leave;
       		return -1;
       	}
       	if ( (bread= usb_pipe_read(njb, status_data, 6)) == -1 ) {
       		NJB_ERROR(EO_USBBLK);
       		__leave;
       		return -1;
       	} else if ( bread < 6 ) {
       		NJB_ERROR(EO_RDSHORT);
       		__leave;
       		return -1;
       	}
	status = (status_data[0]<<8) + status_data[1];
	if (status != 0){
       		__leave;
       		return -1;
       	}
	storedsize = (status_data[2]<<24) + (status_data[3]<<16) + (status_data[4]<<8) + status_data[5];
	if (storedsize < chunksize) {
		NJB_ERROR(EO_WRSHORT);
       		__leave;
       		return -1;
       	}
	return 0;
}

int njb3_send_file_complete(njb_t *njb, u_int32_t fileid)
{
	__dsub= "njb3_send_file_complete";
	/*
	  unsigned char njb3_verify_transfer[]={0x00,0x10,0x00,0x01,0x00,0x00,0x00,0x00,
					      0x00,0x0a,0x00,0x14,0x00,0x00,0x00,0x00,
					      0x00,0x05,0xe8,0x62,0x00,0x00};
	*/
	/* Structure: 22 bytes
       	 * 2 bytes unknown 0x0010
       	 * 2 bytes unknown 0x0001
	 * 18 bytes unknown stuff
	 * I suspect this is a CRC sum check for the transfer. We
	 * don't use it right now, but could be made to work. Reads
	 * back 2 bytes status code, which is OK when its 0x0000 I think.
       	 */
       	unsigned char njb3_send_file_complete[]={0x00,0x09,0x00,0x01,0x00,0x00,0x00,0x00};
	/* Structure: 8 bytes
       	 * 2 bytes unknown 0x0009
       	 * 2 bytes unknown 0x0001
	 * 4 bytes track ID
       	 */
	u_int32_t bread;
	u_int32_t status;
	unsigned char status_data[]={0x00,0x00};
	
	__enter;

	/*
	 * Here, the verification command above should be sent. We
	 * yet don't know how to implement the verification however, someone
	 * has to take a look at the transfers and try to reverse engineer the
	 * algorithm for calculation of the verification command (possibly
	 * a CRC or simple checksum).
	 *
	 * After the verification command, 2 bytes are read in from the bus. 
	 * These should probably be 0x0000 if the checksum/CRC was correct.
	 */

	/* Prepare the command */
	njb3_send_file_complete[4] = (fileid >> 24) & 255;
	njb3_send_file_complete[5] = (fileid >> 16) & 255;
	njb3_send_file_complete[6] = (fileid >> 8) & 255;
	njb3_send_file_complete[7] = fileid & 255;

       	if (send_njb3_command(njb, njb3_send_file_complete, 8) == -1){
       		__leave;
       		return -1;
       	}
       	if ( (bread= usb_pipe_read(njb, status_data, 2)) == -1 ) {
       		NJB_ERROR(EO_USBBLK);
       		__leave;
       		return -1;
       	} else if ( bread < 2 ) {
       		NJB_ERROR(EO_RDSHORT);
       		__leave;
       		return -1;
       	}
	status = (status_data[0]<<8) + status_data[1];
	if (status != 0){
       		__leave;
       		return -1;
       	}

	return 0;
}

int njb3_delete_item (njb_t *njb, u_int32_t itemid)
{
	__dsub= "njb3_delete_item";
       	unsigned char njb3_delitem[]={0x00,0x05,0x00,0x01,0x00,0x00,0x00,0x00};
	/* Structure: 8 bytes
       	 * 2 bytes unknown 0x0005
       	 * 2 bytes unknown 0x0001
	 * 4 bytes item ID (could be track, playlist, datafile ...)
       	 */
	unsigned char status_data[]={0x00,0x00};
	ssize_t bread;
	u_int32_t status;

	__enter;

	njb3_delitem[4] = (itemid>>24) & 255;
	njb3_delitem[5] = (itemid>>16) & 255;
	njb3_delitem[6] = (itemid>>8) & 255;
	njb3_delitem[7] = itemid & 255;

       	if (send_njb3_command(njb, njb3_delitem, 0x08) == -1){
       		__leave;
       		return -1;
       	}
       	if ( (bread= usb_pipe_read(njb, status_data, 2)) == -1 ) {
       		NJB_ERROR(EO_USBBLK);
       		__leave;
       		return -1;
       	} else if ( bread < 2 ) {
       		NJB_ERROR(EO_RDSHORT);
       		__leave;
       		return -1;
       	}
	status = (status_data[0]<<8) + status_data[1];
	if (status) {
       		__leave;
       		return -1;
	}

	__leave;
	return 0;
}

int njb3_update_16bit_frame(njb_t *njb, u_int32_t itemid, u_int32_t frameid, u_int32_t value)
{
	__dsub= "njb3_update_16bit_frame";
       	unsigned char njb3_update16[]={0x00,0x01,0x00,0x01,0x00,0x00,0x00,0x00,
				       0x00,0x04,0x00,0x00,0x00,0x00,0x00,0x00};
	/* Structure: 8 bytes
       	 * 2 bytes unknown 0x0001
       	 * 2 bytes unknown 0x0001
	 * 4 bytes track ID
	 * 2 bytes tag frame length (in this case 4)
	 * 2 bytes tag frame ID
	 * 2 bytes tag frame content
	 * 2 bytes termination 0x0000
       	 */
	unsigned char status_data[]={0x00,0x00};
	ssize_t bread;
	u_int32_t status;

	__enter;

	njb3_update16[4] = (itemid>>24) & 255;
	njb3_update16[5] = (itemid>>16) & 255;
	njb3_update16[6] = (itemid>>8) & 255;
	njb3_update16[7] = itemid & 255;

	njb3_update16[10] = (frameid>>8) & 255;
	njb3_update16[11] = frameid & 255;

	njb3_update16[12] = (value>>8) & 255;
	njb3_update16[13] = value & 255;

       	if (send_njb3_command(njb, njb3_update16, 0x10) == -1){
       		__leave;
       		return -1;
       	}
       	if ( (bread= usb_pipe_read(njb, status_data, 2)) == -1 ) {
       		NJB_ERROR(EO_USBBLK);
       		__leave;
       		return -1;
       	} else if ( bread < 2 ) {
       		NJB_ERROR(EO_RDSHORT);
       		__leave;
       		return -1;
       	}
	status = (status_data[0]<<8) + status_data[1];
	if (status) {
       		__leave;
       		return -1;
	}

	__leave;
	return 0;
}

int njb3_update_string_frame(njb_t *njb, u_int32_t itemid, u_int32_t frameid, unsigned char *str)
{
	__dsub= "njb3_update_string_frame";
       	unsigned char njb3_update_frame[]={0x00,0x01,0x00,0x01,0x00,0x00,0x00,0x00};
	/* Structure: 8 bytes
       	 * 2 bytes unknown 0x0001
       	 * 2 bytes unknown 0x0001
	 * 4 bytes track ID
	 * 2 bytes tag frame length (in this case 4)
	 * 2 bytes tag frame ID
	 * 2 bytes tag frame content
	 * 2 bytes termination 0x0000
       	 */
	unsigned char status_data[]={0x00,0x00};
	unsigned char *data;
	ssize_t bread;
	u_int32_t status;
	u_int32_t strsize;
	u_int32_t framesize;
	u_int32_t cmdsize;

	__enter;

	njb3_update_frame[4] = (itemid>>24) & 255;
	njb3_update_frame[5] = (itemid>>16) & 255;
	njb3_update_frame[6] = (itemid>>8) & 255;
	njb3_update_frame[7] = itemid & 255;

	strsize = ucs2strlen(str)*2 + 2;
	framesize = strsize + 2;
	cmdsize =  12 + framesize;

	data= malloc(cmdsize);
	if (data == NULL) {
		NJB_ERROR(EO_NOMEM);
		__leave;
		return 0;
	}
	memset(data, 0, cmdsize);
	memcpy(&data[0], njb3_update_frame, 8);
	data[8] = (framesize >> 8) & 255;
	data[9] = framesize & 255;
	data[10] = (frameid >> 8) & 255;
	data[11] = frameid & 255;
	memcpy(&data[12], str, strsize);

	if (send_njb3_command(njb, data, cmdsize) == -1){
		free(data);
       		__leave;
       		return -1;
       	}
       	if ( (bread= usb_pipe_read(njb, status_data, 2)) == -1 ) {
		free(data);
       		NJB_ERROR(EO_USBBLK);
       		__leave;
       		return -1;
       	} else if ( bread < 2 ) {
		free(data);
       		NJB_ERROR(EO_RDSHORT);
       		__leave;
       		return -1;
       	}
	status = (status_data[0]<<8) + status_data[1];
	if (status) {
		free(data);
       		__leave;
       		return -1;
	}

	free(data);
	__leave;
	return 0;
}

int njb3_create_playlist (njb_t *njb, char *name, u_int32_t *plid)
{
	__dsub= "njb3_create_playlist";
       	unsigned char njb3_create_pl[]={0x00,0x0a,0x00,0x01,0x00,0x00,0x00,0x01};
	/* Structure: 8 bytes
       	 * 2 bytes unknown 0x000a
       	 * 2 bytes unknown 0x0001
       	 * 2 bytes unknown 0x0000
       	 * 2 bytes unknown 0x0001 (number of tag frames following?)
       	 */
	unsigned char status_data[]={0x00,0x00,0x00,0x00,0x00,0x00};
	unsigned char *data;
	ssize_t bread;
	u_int32_t status;
	u_int32_t strsize;
	u_int32_t framesize;
	u_int32_t cmdsize;

	__enter;

	strsize = ucs2strlen(name)*2 + 2;
	framesize = strsize + 2;
	cmdsize =  16 + framesize;

	data= malloc(cmdsize);
	if (data == NULL) {
		NJB_ERROR(EO_NOMEM);
		__leave;
		return -1;
	}
	memset(data, 0, cmdsize);
	memcpy(&data[0], njb3_create_pl, 8);

	data[8] = (framesize >> 8) & 255;
	data[9] = framesize & 255;
	data[10] = (NJB3_PLNAME_FRAME_ID >> 8) & 255;
	data[11] = NJB3_PLNAME_FRAME_ID & 255;

	memcpy(&data[12], name, strsize);

	if (send_njb3_command(njb, data, cmdsize) == -1){
		free(data);
       		__leave;
       		return -1;
       	}
       	if ( (bread= usb_pipe_read(njb, status_data, 6)) == -1 ) {
		free(data);
       		NJB_ERROR(EO_USBBLK);
       		__leave;
       		return -1;
       	} else if ( bread < 6 ) {
		free(data);
       		NJB_ERROR(EO_RDSHORT);
       		__leave;
       		return -1;
       	}
	status = (status_data[0]<<8) + status_data[1];
	if (status) {
		free(data);
       		__leave;
       		return -1;
	}
	/* Return the new playlist ID */
	*plid = (status_data[2]<<24) + (status_data[3]<<16) + (status_data[4]<<8) + status_data[5];

	free(data);
	__leave;
	return 0;
}

int njb3_add_multiple_tracks_to_playlist (njb_t *njb, u_int32_t plid, u_int32_t *trids, u_int16_t ntracks)
{
	__dsub= "njb3_add_multiple_tracks_to_playlist";
       	unsigned char njb3_addtracks[]={0x01,0x07,0x00,0x01,0x00,0x00,0x00,0x00,0x00,0x00,0x01,0x1c};
	/* Structure: 0x0e + n*2 bytes
       	 * 2 bytes unknown 0x0107
       	 * 2 bytes unknown 0x0001
	 * 4 bytes playlist ID
	 * 2 bytes ntracks = following tracklist length
	 * 2 bytes 0x011c NJB3_PLTRACKS_FRAME_ID
	 * ntracks * 4 bytes track IDs of tracks to add.
	 * 2 bytes unknown 0x0000
       	 */
	unsigned char status_data[]={0x00,0x00,0x00,0x00,0x00,0x00};
	unsigned char *data;
	u_int32_t trackcmdsize;
	u_int32_t cmdsize;
	u_int32_t bp;
	u_int32_t newplid;
	ssize_t bread;
	u_int32_t status;
	int i;

	__enter;
	trackcmdsize = ntracks*4 + 2;
	cmdsize = 0x0c + trackcmdsize;


	data= malloc(cmdsize);
	if (data == NULL) {
		NJB_ERROR(EO_NOMEM);
		__leave;
		return -1;
	}
	memset(data, 0, cmdsize);
	memcpy(&data[0], njb3_addtracks, 12);
	/* add playlist ID */
	data[4] = (plid >> 24) & 255;
	data[5] = (plid >> 16) & 255;
	data[6] = (plid >> 8) & 255;
	data[7] = plid & 255;
	/* Add number of tracks */
	data[8] = (trackcmdsize >> 8) & 255;
	data[9] = trackcmdsize & 255;
	/* Add the track IDs */
	bp = 0;
	for (i= 0; i< ntracks; i++) {
		data[12+bp] = (trids[i] >> 24) & 255;
		data[13+bp] = (trids[i] >> 16) & 255;
		data[14+bp] = (trids[i] >> 8) & 255;
		data[15+bp] = trids[i] & 255;
		bp+= 4;
	}
	/* Send the command */
	if (send_njb3_command(njb, data, cmdsize) == -1){
		free(data);
       		__leave;
       		return -1;
       	}
       	if ( (bread= usb_pipe_read(njb, status_data, 6)) == -1 ) {
		free(data);
       		NJB_ERROR(EO_USBBLK);
       		__leave;
       		return -1;
       	} else if ( bread < 6 ) {
		free(data);
       		NJB_ERROR(EO_RDSHORT);
       		__leave;
       		return -1;
       	}
	status = (status_data[0]<<8) + status_data[1];
	if (status) {
		free(data);
       		__leave;
       		return -1;
	}
	/* Retrieve the new playlist ID */
	newplid = (status_data[2]<<24) + (status_data[3]<<16) + (status_data[4]<<8) + status_data[5];

	/* newplid != plid */
	/* So the user program obviously has to rescan the playlists
	 * as the new playlist ID is not returned ... Perhaps not
	 * so good. */

	__leave;
	return 0;
}

int njb3_adjust_sound(njb_t *njb, u_int16_t effect, u_int16_t value)
{
  __dsub= "njb3_adjust_sound";
  unsigned char njb3_adjust[]={0x02, 0x01, 0x00, 0x01, 0x02, 0x03, 0x00, 0x00,
			       0x00, 0x04, 0x02, 0x02, 0x00, 0x01, 0x00, 0x04,
			       0x02, 0x03, 0x00, 0x49, 0x00, 0x00};
  /* Structure: 0x16 bytes
   * 2 bytes unknown 0x0201
   * 2 bytes unknown 0x0001
   * 2 bytes unknown 0x0203
   * 2 bytes termination 0x0000
   * 2 bytes length 0x0004
   * 2 bytes unknown frame ID    0x0202 
   * 2 bytes unknown frame value 0x0001
   * 2 bytes length              0x0004
   * 2 bytes parameter frame ID  0x0203
   * 2 bytes volume
   * 2 bytes termination 0x0000
   */
  unsigned char status[2];
  int bread;

  __enter;

  /* Effect to be adjusted */
  from_16bit_to_njb3_bytes(effect, &njb3_adjust[0x10]);
  /* Value */
  from_16bit_to_njb3_bytes(value, &njb3_adjust[0x12]);
  if (send_njb3_command(njb, njb3_adjust, 0x16) == -1){
    __leave;
    return -1;
  }

  /* Read back status (command result) */
  if ( (bread= usb_pipe_read (njb, &status, 2)) == -1 ) {
    NJB_ERROR (EO_USBBLK);
    __leave;
    return -1;
  }
  else if (bread < 2) {
    NJB_ERROR (EO_RDSHORT);
    __leave;
    return -1;
  }
  if (njb3_bytes_to_16bit(&status[0]) != 0) {
    __leave;
    return -1;
  }

  __leave;

  return 0;
}

eax_t *njb3_get_eax (njb_t *njb)
{
  unsigned char *data = NULL;

  return eax_unpack3(data);
}
