#ifndef __NJB__PROTO3__H
#define __NJB__PROTO3__H

/* Buffer for short reads */
#define NJB3_SHORTREAD_BUFSIZE 1024
#define NJB3_METADATA_BUFSIZE 0x10010

/* These are the metadata frame IDs used by NJB3 */
#define NJB3_UNKNOWN1_FRAME_ID  0x0001 /* 8 bytes - requested at startup */
#define NJB3_DISKUSAGE_FRAME_ID 0x0002 /* 14 bytes */
#define NJB3_PRODID_FRAME_ID    0x0003 /* String - product ID */
#define NJB3_LOCKED_FRAME_ID    0x0006 /* 16 bit word */
#define NJB3_FNAME_FRAME_ID     0x0007 /* String: Original filename on host */
#define NJB3_UNKNOWN4_FRAME_ID  0x000a /* 4 bytes string "AR00", "PL00", "SG00" known */
#define NJB3_CODEC_FRAME_ID     0x000b /* 16 bit word */
#define NJB3_POSTID_FRAME_ID    0x000c /* 16 bit word TrackID on tracks, PlaylistID on playlists */
#define NJB3_DIR_FRAME_ID       0x000d /* String: Original directory on host */
#define NJB3_FILESIZE_FRAME_ID  0x000e /* 32 bit word */
#define NJB3_FILECOUNT_FRAME_ID 0x0013 /* 6 bytes: at least top 2 bytes are file & directory count */
#define NJB3_UNKNOWN5_FRAME_ID  0x0014 /* 8 bytes, 2*32 bit words */
#define NJB3_JUKEBOXID_FRAME_ID 0x0015 /* 18 bytes - unique device ID */
#define NJB3_FILETIME_FRAME_ID  0x0016 /* 32 bit word timestamp (UNIX format) */
#define NJB3_UNKNOWN6_FRAME_ID  0x0017 /* 32 bit word */
#define NJB3_UNKNOWN8_FRAME_ID  0x0018 /* 32 bit word set on files, often 0x2000 0x0000 */
#define NJB3_ALBUM_FRAME_ID     0x0101 /* String */
#define NJB3_ARTIST_FRAME_ID    0x0102 /* String */
#define NJB3_GENRE_FRAME_ID     0x0103 /* String */
#define NJB3_TITLE_FRAME_ID     0x0104 /* String */
#define NJB3_LENGTH_FRAME_ID    0x0105 /* String */
#define NJB3_TRACKNO_FRAME_ID   0x0106 /* 16 bit word */
#define NJB3_YEAR_FRAME_ID      0x0107 /* 16 bit word */
#define NJB3_TRACKINFO_FRAME_ID 0x010b /* 16 bit word: Info about current track on the playlist */
#define NJB3_SEEKTRACK_FRAME_ID 0x010c /* 32 bit word: seek to position (in ms) in current track */
#define NJB3_PLNAME_FRAME_ID    0x010f /* String: playlist name */
#define NJB3_TIME_FRAME_ID      0x0110 /* 10 bytes */
#define NJB3_OWNER_FRAME_ID     0x0113 /* String: owner name */
#define NJB3_UNKNOWN3_FRAME_ID  0x0115 /* 32bit word - unknown use,
					 requested after reading tracks value on NJB3 0x0000 0x0003 */
#define NJB3_PLAYTRACK_FRAME_ID 0x0119 /* 32 bit word track id to play */
#define NJB3_PLTRACKS_FRAME_ID  0x011c /* Array of 16bit words */
#define NJB3_VOLUME_FRAME_ID    0x0203 /* 16 bit word */
#define NJB3_SPAT_FRAME_ID      0x0206 /* 16 bit word spatialization 2 = full */
#define NJB3_UNKNOWN7_FRAME_ID  0x1400 /* Array of AR00, PL00 and SG00 metadatas, requesed in a
					  subrequest parameter to this request */

/* File type IDs, used for eg create_file */
#define NJB3_FILETYPE_FILE     0x0000
#define NJB3_FILETYPE_TRACK    0x0002

/* Codec IDs */
#define NJB3_CODEC_MP3_ID            0x0000
#define NJB3_CODEC_WAV_ID            0x0001
#define NJB3_CODEC_PRELOADED_MP3_ID  0x0002
#define NJB3_CODEC_WMA_ID            0x0003 /* Wild guess */

/* Transfer block size (adding and fetching tracks, files */
#define NJB3_GET_FILE_BLOCK_SIZE    0x2000
#define NJB3_SEND_FILE_BLOCK_SIZE   0x2000

/* Stop, pause and resume are very much alike. */
#define NJB3_START_PLAY     0x00
#define NJB3_STOP_PLAY      0x01
#define NJB3_PAUSE_PLAY     0x02
#define NJB3_RESUME_PLAY    0x03

#define njb3_start_play(njb) njb3_ctrl_playing(njb, NJB3_START_PLAY)
#define njb3_stop_play(njb) njb3_ctrl_playing(njb, NJB3_STOP_PLAY)
#define njb3_pause_play(njb) njb3_ctrl_playing(njb, NJB3_PAUSE_PLAY)
#define njb3_resume_play(njb) njb3_ctrl_playing(njb, NJB3_RESUME_PLAY)

/* NJB3 functions */
int njb2_set_bitmap (njb_t *njb, unsigned char * bitmap);

int njb3_current_track (njb_t *njb, u_int16_t * track);
int njb3_elapsed_time (njb_t *njb, u_int16_t * elapsed, int * change);
int njb3_play_track (njb_t *njb, u_int32_t trackid);
int njb3_queue_track (njb_t *njb, u_int32_t trackid);
int njb3_ctrl_playing (njb_t *njb, int cmd);
int njb3_seek_track (njb_t *njb, u_int32_t position);

int njb3_read_unknown_1(njb_t *njb);
int njb3_read_ar00pl00sg00(njb_t *njb);
int njb3_ping (njb_t *njb, njbid_t *njbid, int type);
int njb3_readid (njb_t *njb, njbid_t *njbid);
/* int njb3_capture (njb_t *njb, int which); */
int njb3_capture (njb_t *njb);
int njb3_release (njb_t *njb);
int njb3_get_disk_usage (njb_t *njb, u_int64_t *totalbytes, u_int64_t *freebytes);
int njb3_get_owner_string (njb_t *njb, char *name);
int njb3_set_owner_string (njb_t *njb, const char *name);
njb_time_t *njb3_get_time(njb_t *njb);
int njb3_set_time(njb_t *njb, njb_time_t *time);
songid_t *njb3_get_first_track_tag (njb_t *njb);
songid_t *njb3_get_next_track_tag (njb_t *njb);
playlist_t *njb3_get_first_playlist_tag (njb_t *njb);
playlist_t *njb3_get_next_playlist_tag (njb_t *njb);
datafile_t *njb3_get_first_datafile_tag (njb_t *njb);
datafile_t *njb3_get_next_datafile_tag (njb_t *njb);
int njb3_request_file(njb_t *njb, u_int32_t fileid, u_int32_t size);
int njb3_get_file_chunk(njb_t *njb, int fd, u_int32_t size);
int njb3_get_file_confirmation(njb_t *njb);
u_int32_t njb3_create_file(njb_t *njb, unsigned char *ptag, u_int32_t tagsize, u_int16_t filetype);
int njb3_send_file_chunk(njb_t *njb, unsigned char *chunk, u_int32_t chunksize, u_int32_t fileid);
int njb3_send_file_complete(njb_t *njb, u_int32_t fileid);
int njb3_delete_item (njb_t *njb, u_int32_t itemid);
int njb3_update_16bit_frame(njb_t *njb, u_int32_t itemid, u_int32_t frameid, u_int32_t value);
int njb3_update_string_frame(njb_t *njb, u_int32_t itemid, u_int32_t frameid, unsigned char *str);
int njb3_create_playlist (njb_t *njb, char *name, u_int32_t *plid);
int njb3_add_multiple_tracks_to_playlist (njb_t *njb, u_int32_t plid, u_int32_t *trids, u_int16_t ntracks);
int njb3_adjust_sound(njb_t *njb, u_int16_t effect, u_int16_t value);
eax_t *njb3_get_eax (njb_t *njb);

#endif
